@@ -1,125 +1,11 @@
+//! Generic object and objectlist wrappers.
 use crate::{
+    api_resource::ApiResource,
     metadata::{ListMeta, ObjectMeta, TypeMeta},
-    ErrorResponse,
+    resource::Resource,
 };
 use serde::{Deserialize, Serialize};
-use std::fmt::Debug;
-
-/// A raw event returned from a watch query
-///
-/// Note that a watch query returns many of these as newline separated JSON.
-#[derive(Deserialize, Serialize, Clone)]
-#[serde(tag = "type", content = "object", rename_all = "UPPERCASE")]
-pub enum WatchEvent<K> {
-    /// Resource was added
-    Added(K),
-    /// Resource was modified
-    Modified(K),
-    /// Resource was deleted
-    Deleted(K),
-    /// Resource bookmark. `Bookmark` is a slimmed down `K` due to [#285](https://github.com/clux/kube-rs/issues/285).
-    ///
-    /// From [Watch bookmarks](https://kubernetes.io/docs/reference/using-api/api-concepts/#watch-bookmarks).
-    ///
-    /// NB: This became Beta first in Kubernetes 1.16.
-    Bookmark(Bookmark),
-    /// There was some kind of error
-    Error(ErrorResponse),
-}
-
-impl<K> Debug for WatchEvent<K> {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match &self {
-            WatchEvent::Added(_) => write!(f, "Added event"),
-            WatchEvent::Modified(_) => write!(f, "Modified event"),
-            WatchEvent::Deleted(_) => write!(f, "Deleted event"),
-            WatchEvent::Bookmark(_) => write!(f, "Bookmark event"),
-            WatchEvent::Error(e) => write!(f, "Error event: {:?}", e),
-        }
-    }
-}
-
-/// Slimed down K for [`WatchEvent::Bookmark`] due to [#285](https://github.com/clux/kube-rs/issues/285).
-///
-/// Can only be relied upon to have metadata with resource version.
-/// Bookmarks contain apiVersion + kind + basically empty metadata.
-#[derive(Serialize, Deserialize, Clone)]
-pub struct Bookmark {
-    /// apiVersion + kind
-    #[serde(flatten)]
-    pub types: TypeMeta,
-
-    /// Basically empty metadata
-    pub metadata: BookmarkMeta,
-}
-
-/// Slimed down Metadata for WatchEvent::Bookmark
-#[derive(Serialize, Deserialize, Clone)]
-#[serde(rename_all = "camelCase")]
-pub struct BookmarkMeta {
-    pub resource_version: String,
-}
-
-// -------------------------------------------------------
-
-/// A standard Kubernetes object with `.spec` and `.status`.
-///
-/// This is a convenience struct provided for serialization/deserialization
-/// It is not useful within the library anymore, because it can not easily implement
-/// the [`k8s_openapi`] traits.
-///
-/// This is what Kubernetes maintainers tell you the world looks like.
-/// It's.. generally true.
-#[derive(Deserialize, Serialize, Clone)]
-pub struct Object<P, U>
-where
-    P: Clone,
-    U: Clone,
-{
-    /// The types field of an `Object`
-    #[serde(flatten)]
-    pub types: TypeMeta,
-
-    /// Resource metadata
-    ///
-    /// Contains information common to most resources about the Resource,
-    /// including the object name, annotations, labels and more.
-    pub metadata: ObjectMeta,
-
-    /// The Spec struct of a resource. I.e. `PodSpec`, `DeploymentSpec`, etc.
-    ///
-    /// This defines the desired state of the Resource as specified by the user.
-    pub spec: P,
-
-    /// The Status of a resource. I.e. `PodStatus`, `DeploymentStatus`, etc.
-    ///
-    /// This publishes the state of the Resource as observed by the controller.
-    /// Use `U = NotUsed` when a status does not exist.
-    #[serde(default, skip_serializing_if = "Option::is_none")]
-    pub status: Option<U>,
-}
-
-impl<P, U> Object<P, U>
-where
-    P: Clone,
-    U: Clone,
-{
-    /// A constructor like the one from kube-derive
-    pub fn new<K: k8s_openapi::Resource>(name: &str, spec: P) -> Self {
-        Self {
-            types: TypeMeta {
-                api_version: <K as k8s_openapi::Resource>::API_VERSION.to_string(),
-                kind: <K as k8s_openapi::Resource>::KIND.to_string(),
-            },
-            metadata: ObjectMeta {
-                name: Some(name.to_string()),
-                ..Default::default()
-            },
-            spec,
-            status: None,
-        }
-    }
-}
+use std::borrow::Cow;
 
 /// A generic Kubernetes object list
 ///
@@ -218,9 +104,138 @@ impl<'a, T: Clone> IntoIterator for &'a mut ObjectList<T> {
 
 // -------------------------------------------------------
 
+/// A standard Kubernetes object with `.spec` and `.status`.
+///
+/// This is a convenience struct provided for serialization/deserialization.
+/// It is slightly stricter than ['DynamicObject`] in that it enforces the spec/status convention,
+/// and as such will not in general work with all api-discovered resources.
+///
+/// This can be used to tie existing resources to smaller, local struct variants to optimize for memory use.
+/// E.g. if you are only interested in a few fields, but you store tons of them in memory with reflectors.
+#[derive(Deserialize, Serialize, Clone, Debug)]
+pub struct Object<P, U>
+where
+    P: Clone,
+    U: Clone,
+{
+    /// The type fields, not always present
+    #[serde(flatten, default)]
+    pub types: Option<TypeMeta>,
+
+    /// Resource metadata
+    ///
+    /// Contains information common to most resources about the Resource,
+    /// including the object name, annotations, labels and more.
+    pub metadata: ObjectMeta,
+
+    /// The Spec struct of a resource. I.e. `PodSpec`, `DeploymentSpec`, etc.
+    ///
+    /// This defines the desired state of the Resource as specified by the user.
+    pub spec: P,
+
+    /// The Status of a resource. I.e. `PodStatus`, `DeploymentStatus`, etc.
+    ///
+    /// This publishes the state of the Resource as observed by the controller.
+    /// Use `U = NotUsed` when a status does not exist.
+    #[serde(default, skip_serializing_if = "Option::is_none")]
+    pub status: Option<U>,
+}
+
+impl<P, U> Object<P, U>
+where
+    P: Clone,
+    U: Clone,
+{
+    /// A constructor that takes Resource values from an `ApiResource`
+    pub fn new(name: &str, ar: &ApiResource, spec: P) -> Self {
+        Self {
+            types: Some(TypeMeta {
+                api_version: ar.api_version.clone(),
+                kind: ar.kind.clone(),
+            }),
+            metadata: ObjectMeta {
+                name: Some(name.to_string()),
+                ..Default::default()
+            },
+            spec,
+            status: None,
+        }
+    }
+
+    /// Attach a namespace to an Object
+    pub fn within(mut self, ns: &str) -> Self {
+        self.metadata.namespace = Some(ns.into());
+        self
+    }
+}
+
+impl<P, U> Resource for Object<P, U>
+where
+    P: Clone,
+    U: Clone,
+{
+    type DynamicType = ApiResource;
+
+    fn group(dt: &ApiResource) -> Cow<'_, str> {
+        dt.group.as_str().into()
+    }
+
+    fn version(dt: &ApiResource) -> Cow<'_, str> {
+        dt.version.as_str().into()
+    }
+
+    fn kind(dt: &ApiResource) -> Cow<'_, str> {
+        dt.kind.as_str().into()
+    }
+
+    fn api_version(dt: &ApiResource) -> Cow<'_, str> {
+        dt.api_version.as_str().into()
+    }
+
+    fn meta(&self) -> &ObjectMeta {
+        &self.metadata
+    }
+
+    fn meta_mut(&mut self) -> &mut ObjectMeta {
+        &mut self.metadata
+    }
+}
+
 /// Empty struct for when data should be discarded
 ///
 /// Not using [`()`](https://doc.rust-lang.org/stable/std/primitive.unit.html), because serde's
 /// [`Deserialize`](serde::Deserialize) `impl` is too strict.
 #[derive(Clone, Deserialize, Serialize, Default, Debug)]
 pub struct NotUsed {}
+
+
+#[cfg(test)]
+mod test {
+    use super::{ApiResource, NotUsed, Object};
+    #[test]
+    fn simplified_k8s_object() {
+        use k8s_openapi::api::core::v1::Pod;
+        // Replacing heavy type k8s_openapi::api::core::v1::PodSpec with:
+        #[derive(Clone)]
+        struct PodSpecSimple {
+            containers: Vec<ContainerSimple>,
+        }
+        #[derive(Clone)]
+        struct ContainerSimple {
+            image: String,
+        }
+        type PodSimple = Object<PodSpecSimple, NotUsed>;
+        // by grabbing the ApiResource info from the Resource trait
+        let ar = ApiResource::erase::<Pod>(&());
+        assert_eq!(ar.group, "");
+        assert_eq!(ar.kind, "Pod");
+        let data = PodSpecSimple {
+            containers: vec![ContainerSimple { image: "blog".into() }],
+        };
+        let mypod = PodSimple::new("blog", &ar, data).within("dev");
+        assert_eq!(mypod.metadata.namespace.unwrap(), "dev");
+        assert_eq!(mypod.metadata.name.unwrap(), "blog");
+        assert_eq!(mypod.types.as_ref().unwrap().kind, "Pod");
+        assert_eq!(mypod.types.as_ref().unwrap().api_version, "v1");
+    }
+}
